""" Handles the GPT API and the conversation state. """
import json
import time
from pathlib import Path
from typing import Optional, Union, List, Callable

from PIL.Image import Image

from justai.models.basemodel import ImageInput
from justai.tools.cache import cached_response, cache_save
from justai.model.message import Message
from justai.models.modelfactory import ModelFactory
from justai.tools.images import crop_to_fit


class Model:
    def __init__(self, model_name: str, **kwargs):
        
        # Model parameters
        self.model = ModelFactory.create(model_name, **kwargs)
        self.model.encapsulating_model = self

        # Parameters to save the current conversation
        self.save_dir = Path(__file__).resolve().parent / 'saves'
        self.message_memory = 20  # Number of messages to remember. Limits token usage.
        self.messages = []  # List of Message objects
        self.tools = []  # List of tools to use / functions to call
        self.functions = {}  # The actual functions to call with key the name of the function and as value the function

        self.input_token_count = 0
        self.output_token_count = 0
        self.last_response_time = 0
        
        self.logger = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """ Closes any open connections in the underlying model. """
        if hasattr(self.model, 'close'):
            self.model.close()

    def __setattr__(self, name, value):
        if name not in self.__dict__ and hasattr(self, 'model') and name in self.model.model_params:
            # Not an existing property model but a model_params property. Set it in model_params
            self.model.model_params[name] = value
        else:
            # Update the property as intended
            super().__setattr__(name, value)

    def set_api_key(self, key: str):
        """ Used when using Aigent from a browser where the user has to specify a key """
        self.model.set('api_key', key)

    @property
    def model_name(self):
        return self.model.model_name

    @property
    def system(self):  # This function can be overwritten by child classes to make the system message dynamic
        return self.model.system_message

    @system.setter
    def system(self, value):
        self.model.system_message = value

    @property
    def cached_prompt(self): 
        if hasattr(self.model, 'cached_prompt'):
            return self.model.cached_prompt
        raise AttributeError("Model does not support cached_prompt")

    @cached_prompt.setter
    def cached_prompt(self, value):
        if hasattr(self.model, 'cached_prompt'):
            self.model.cached_prompt = value
        else:
            raise AttributeError("Model does not support cached_prompt")

    @property
    def cache_creation_input_tokens(self):
        if hasattr(self.model, 'cache_creation_input_tokens'):
            return self.model.cache_creation_input_tokens
        raise AttributeError("Model does not support cache_creation_input_tokens")
    
    @property
    def cache_read_input_tokens(self):
        if hasattr(self.model, 'cache_read_input_tokens'):
            return self.model.cache_read_input_tokens
        raise AttributeError("Model does not support cache_read_input_tokens")
        
    def reset(self):
        self.messages = []

    def add_tool(self, function: Callable, description: str|None = None, parameters: dict = None, required_parameters: list=None):
        if description is None and not self.model.supports_function_calling:
            raise NotImplementedError(f"{self.model.model_name} does not support function calling")
        if not description and not self.model.supports_automatic_function_calling:
            raise NotImplementedError(f"{self.model.model_name} does not support automatic function calling")
        tool = {'type': 'function', 'function': function, 'description': description, 'parameters': parameters,
                'required_parameters': required_parameters}
        self.tools.append(tool)
        self.functions[function.__name__] = function

    def last_token_count(self):
        return self.input_token_count, self.output_token_count, self.input_token_count + self.output_token_count

    def prompt(self, prompt: str, *, images: ImageInput = None, return_json=False, response_format=None, cached=True):
        self.raise_for_unsupported()

        start_time = time.time()
        if images and not isinstance(images, list):
            images = [images]

        response = None

        if cached:
            response = cached_response(self.model.model_name, self.model.model_params, prompt, images, self.tools,
                                       return_json, response_format)

        if response:
            result, _, _, tool_use = response
            self.input_token_count = self.output_token_count = 0
        else:
            response = self.model.prompt(prompt, images=images, tools=self.tools, return_json=return_json,
                                         response_format=response_format)
            if cached:
                cache_save(response, self.model.model_name, self.model.model_params, prompt, images, self.tools,
                           return_json, response_format)

            result, self.input_token_count, self.output_token_count = response

        self.last_response_time = time.time() - start_time
        return result

    def chat(self, prompt: str, *, images: ImageInput = None, return_json=False, response_format=None, cached=False):
        self.raise_for_unsupported()
        if cached:
            raise NotImplementedError("Model.chat does not support cached=True. Use prompt instead.")

        start_time = time.time()
        if images and not isinstance(images, list):
            images = [images]

        response = self.model.chat(prompt, images=images, tools=self.tools, return_json=return_json,
                                   response_format=response_format)

        result, self.input_token_count, self.output_token_count = response
        self.last_response_time = time.time() - start_time
        return result
    
    async def prompt_async(self, prompt, *, images: ImageInput = None):
        # Using 'async for' to properly yield from the chat_async generator
        async for content, reasoning in self.model.prompt_async(prompt=prompt):
            yield content, reasoning

    async def chat_async(self, prompt, *, images: ImageInput = None):
        if images and not isinstance(images, list):
            images = [images]
        for word in self.model.chat_async(prompt=prompt, images=images):
            if word:
                yield word

    async def prompt_async_reasoning(self, prompt, *, images: ImageInput = None):
        self.reset()
        # Using 'async for' to properly yield from the chat_async_reasoning generator
        async for word, reasoning_content in self.chat_async_reasoning(prompt=prompt, images=images):
            yield word, reasoning_content

    async def chat_async_reasoning(self, prompt, *, images: ImageInput = None):
        """ Same as chat_async but returns the reasoning content as well
        """
        if images and not isinstance(images, list):
            images = [images]
        async for word, reasoning_content in self.model.chat_async(prompt=prompt, images=images):
            if word or reasoning_content:
                yield word, reasoning_content

    def raise_for_unsupported(self, images: ImageInput = None, return_json=False):
        if return_json and not self.model.supports_return_json:
            raise NotImplementedError(f"{self.model.model_name} does not support return_json")
        if images and not self.model.supports_image_input:
            raise NotImplementedError(f"{self.model.model_name} does not support image input")

    def token_count(self, text: str):
        return self.model.token_count(text)

    def generate_image(self, prompt: str, images: ImageInput = None, size: tuple[int, int]|None = None, options: dict = None) -> Image:
        if not self.model.supports_image_generation:
            raise NotImplementedError(f"{self.model.model_name} does not support image generation")
        if images and not isinstance(images, list):
            images = [images]
        image = self.model.generate_image(prompt, images, options=options)
        if size:
            image = crop_to_fit(image, size[0], size[1])
        return image




""" Implementation of the Anthropic models.

Feature table:
    - Async chat:       YES (1)
    - Return JSON:      YES (2)
    - Structured types: YES (Claude 4.x only)
    - Token count:      YES
    - Image support:    YES
    - Tool use:         YES

Models:
Claude 3 Opus:   claude-3-opus-20240229
Claude 3 Sonnet: claude-3-5-sonnet-20240620
Claude 3 Haiku:  claude-3-haiku-20240307
Claude 4 Sonnet: claude-sonnet-4-5-*
Claude 4 Opus:   claude-opus-4-*
Claude 4 Haiku:  claude-haiku-4-*

Supported parameters:
max_tokens: int (default 800)
temperature: float (default 0.8)

(1) In contrast to Model.chat, Model.chat_async cannot return json and does not return input and output token counts
(2) Claude 4.x uses structured outputs beta for guaranteed JSON. Claude 3.x uses legacy parsing with deprecation warning.

"""

import json
import logging
import os
import re
from typing import Any, AsyncGenerator

import httpx
from anthropic import Anthropic, AsyncAnthropic, APIConnectionError, AuthenticationError, PermissionDeniedError, \
    APITimeoutError, RateLimitError, BadRequestError, InternalServerError
from dotenv import dotenv_values

logger = logging.getLogger(__name__)

# Models that support structured outputs (beta)
# See: https://platform.claude.com/docs/en/build-with-claude/structured-outputs
STRUCTURED_OUTPUT_MODELS = re.compile(r'claude-(sonnet-4|opus-4|haiku-4)')
STRUCTURED_OUTPUTS_BETA = 'structured-outputs-2025-11-13'


from justai.model.message import Message
from justai.models.basemodel import (
    BaseModel,
    DEFAULT_TIMEOUT,
    identify_image_format_from_base64,
    ConnectionException,
    AuthorizationException,
    ModelOverloadException,
    RatelimitException,
    BadRequestException,
    GeneralException,
    ImageInput,
)
from justai.tools.display import ERROR_COLOR, color_print
from justai.tools.images import to_base64_image


class AnthropicModel(BaseModel):
    def __init__(self, model_name: str, params: dict = None):
        params = params or {}
        system_message = f"You are {model_name}, a large language model trained by Anthropic."
        super().__init__(model_name, params, system_message)
        self.cached_prompt = None

        # Authentication
        if "ANTHROPIC_API_KEY" in params:
            api_key = params["ANTHROPIC_API_KEY"]
            del params["ANTHROPIC_API_KEY"]
        else:
            api_key = os.getenv("ANTHROPIC_API_KEY") or dotenv_values()["ANTHROPIC_API_KEY"]
        if not api_key:
            color_print(
                "No Anthropic API key found. Create one at https://console.anthropic.com/settings/keys and "
                + "set it in the .env file like ANTHROPIC_API_KEY=here_comes_your_key.",
                color=ERROR_COLOR,
            )

        # Client
        timeout = httpx.Timeout(params.get('timeout', DEFAULT_TIMEOUT))
        if params.get('async'):
            http_client = httpx.AsyncClient(timeout=timeout)
            self.client = AsyncAnthropic(api_key=api_key, http_client=http_client)
        else:
            http_client = httpx.Client(timeout=timeout)
            self.client = Anthropic(api_key=api_key, http_client=http_client)

        # Required model parameters
        if "max_tokens" not in params:
            params["max_tokens"] = 800

        self.supports_cached_prompts = True
        self.messages = []

    def prompt(self, prompt: str, images: ImageInput = None, tools: list = None, return_json: bool = False, response_format=None) \
            -> tuple[Any, int|None, int|None, dict|None]:
        # Reset messages
        self.messages = []
        return self.chat(prompt, images, tools, return_json, response_format)


    def chat(self, prompt: str, images: ImageInput = None, tools: list = None, return_json: bool = False, response_format=None) \
            -> tuple[Any, int|None, int|None, dict|None]:
        # Check if we should use structured outputs (beta) for newer models
        use_structured_outputs = return_json and self._supports_structured_outputs()

        if use_structured_outputs:
            try:
                message = self._completion_with_structured_output(prompt, images, tools, response_format)
            except TypeError as e:
                # Fall back to legacy path if SDK doesn't support structured outputs
                # This happens with older anthropic SDK versions
                logger.warning(
                    f'Structured outputs not available (SDK may be outdated): {e}. '
                    'Falling back to legacy JSON parsing. '
                    'Upgrade anthropic SDK with: pip install --upgrade anthropic'
                )
                use_structured_outputs = False
                message = self.completion(prompt, images, tools, return_json, response_format)
            except BadRequestError as e:
                # Fall back to legacy path if model doesn't support structured outputs
                if 'does not support output format' in str(e):
                    logger.warning(f'Model {self.model_name} does not support structured outputs. Falling back to legacy JSON parsing.')
                    use_structured_outputs = False
                    message = self.completion(prompt, images, tools, return_json, response_format)
                else:
                    raise
        else:
            message = self.completion(prompt, images, tools, return_json, response_format)

        # Text content
        response_str = message.content[0].text
        if return_json:
            if use_structured_outputs:
                # Structured outputs guarantees valid JSON
                response = json.loads(response_str)
            else:
                # Legacy JSON parsing for older models or fallback
                response = self._parse_json_legacy(response_str)
        else:
            response = response_str

        # Token count
        input_tokens = message.usage.input_tokens
        output_tokens = message.usage.output_tokens
        if self.cached_prompt:
            self.cache_creation_input_tokens = message.usage.cache_creation_input_tokens
            self.cache_read_input_tokens = message.usage.cache_read_input_tokens
        else:
            self.cache_creation_input_tokens = self.cache_read_input_tokens = 0

        return response, input_tokens, output_tokens

    def _supports_structured_outputs(self) -> bool:
        """Check if the model supports structured outputs (Claude 4.x models)."""
        return bool(STRUCTURED_OUTPUT_MODELS.search(self.model_name))

    def _parse_json_legacy(self, response_str: str) -> dict | list:
        """
        Legacy JSON parsing for older models that don't support structured outputs.
        TODO: Remove when Claude 3.x models are deprecated.
        """
        response_str = response_str.split('</json>')[0]
        if response_str.startswith('```json'):
            response_str = response_str[7:-3]
        try:
            return json.loads(response_str, strict=False)
        except json.decoder.JSONDecodeError:
            logger.error(f'Error decoding JSON: {response_str[:200]}')
            try:
                return extract_dict(response_str)
            except json.decoder.JSONDecodeError as e:
                raise ValueError(f'Expected JSON but got: {response_str[:200]}') from e

    def _completion_with_structured_output(self, prompt: str, images: ImageInput = None,
                                           tools: list = None, response_format=None):
        """Use the structured outputs beta API for guaranteed JSON responses."""
        system_message = self.cached_system_message() if self.cached_prompt else self.system_message

        # Add user message to conversation history
        if prompt or images:
            user_message = create_anthropic_message('user', prompt, images)
            self.messages.append(user_message)

        # Build output_format schema
        if response_format and hasattr(response_format, 'model_json_schema'):
            # Pydantic model - extract JSON schema
            schema = response_format.model_json_schema()
        elif response_format and isinstance(response_format, dict):
            # Already a JSON schema dict
            schema = response_format
        else:
            # Default: accept any JSON object
            schema = {'type': 'object'}

        # Ensure additionalProperties is set to false (required by Anthropic)
        if schema.get('type') == 'object' and 'additionalProperties' not in schema:
            schema['additionalProperties'] = False

        output_format_param = {
            'type': 'json_schema',
            'schema': schema
        }

        antr_tools = transform_tools(tools or []) if tools is not None else None

        try:
            api_params = {
                'model': self.model_name,
                'messages': self.messages,
                'system': system_message,
                **self.model_params
            }

            if antr_tools:
                api_params['tools'] = antr_tools

            # Use parse() for Pydantic models (recommended by Anthropic SDK)
            if response_format and hasattr(response_format, 'model_json_schema'):
                return self.client.beta.messages.parse(
                    betas=[STRUCTURED_OUTPUTS_BETA],
                    output_format=response_format,
                    **api_params
                )
            else:
                # Use create() with output_format for raw JSON schemas
                return self.client.beta.messages.create(
                    betas=[STRUCTURED_OUTPUTS_BETA],
                    output_format=output_format_param,
                    **api_params
                )

        except TypeError:
            # Re-raise TypeError directly for fallback handling in chat()
            # This happens when SDK doesn't support output_format parameter
            raise
        except APIConnectionError as e:
            logger.error(f'LLM call failed (APIConnectionError): {e!r}')
            raise ConnectionException(e)
        except (AuthenticationError, PermissionDeniedError) as e:
            logger.error(f'LLM call failed (Auth): {e!r}')
            raise AuthorizationException(e)
        except InternalServerError as e:
            logger.error(f'LLM call failed (500): {e!r}')
            raise ModelOverloadException(e)
        except RateLimitError as e:
            logger.error(f'LLM call failed (RateLimit): {e!r}')
            raise RatelimitException(e)
        except BadRequestError as e:
            # Re-raise "does not support output format" errors for fallback handling in chat()
            if 'does not support output format' in str(e):
                raise
            logger.error(f'LLM call failed (BadRequest): {e!r}')
            raise BadRequestException(e)
        except Exception as e:
            logger.error(f'LLM call failed (Unexpected): {e!r}')
            raise GeneralException(e)

    async def prompt_async(self, prompt: str, images: ImageInput = None) -> AsyncGenerator[tuple[str, str], None]:
        async for content, reasoning in self.chat_async(prompt, images):
            if content or reasoning:
                yield content, reasoning

    async def chat_async(self, prompt: str, images: ImageInput = None) -> AsyncGenerator[tuple[str, str], None]:
        stream = self.completion(prompt, images, stream=True)
        for event in stream:
            if hasattr(event, "delta") and hasattr(event.delta, "text"):
                yield event.delta.text, None  # 2nd parameter is reasoning_content. Not available yet for Anthropic

    def completion(self, prompt: str, images: ImageInput = None, tools: list = None, return_json: bool = False, response_format=None, stream=False):
        system_message = self.cached_system_message() if self.cached_prompt else self.system_message
        
        # Add user message to conversation history if it's a new message
        if prompt or images:
            user_message = create_anthropic_message('user', prompt, images)
            self.messages.append(user_message)
        
        antr_tools = transform_tools(tools or []) if tools is not None else None

        for _ in range(3):
            try:
                if stream:
                    if tools:
                        raise NotImplementedError('Anthropic model does not support streaming and tools at the same time')
                    return self.client.messages.create(
                        model=self.model_name,
                        system=system_message,
                        messages=self.messages,
                        stream=True,
                        **self.model_params
                    )
                
                # Prepare messages for the API call
                api_messages = []
                tool_use_context = {}
                
                for msg in self.messages:
                    if msg['role'] == 'user':
                        # For user messages, include them as-is
                        api_messages.append(msg)
                    elif msg['role'] == 'assistant':
                        # For assistant messages, include tool_use blocks and text content
                        content = []
                        for item in msg.get('content', []):
                            if isinstance(item, dict):
                                if item.get('type') == 'tool_use':
                                    # Store tool use context for later reference
                                    tool_use_context[item['id']] = item
                                    content.append(item)
                                elif item.get('type') == 'text':
                                    content.append(item)
                            elif isinstance(item, str):
                                content.append({"type": "text", "text": item})
                        
                        if content:
                            api_messages.append({"role": "assistant", "content": content})
                    
                    # Tool results are handled as part of the next user message
                    # So we don't need to add them to api_messages
                
                # Prepare the API call parameters
                api_params = {
                    'model': self.model_name,
                    'messages': api_messages,
                    **self.model_params
                }
                
                # Only add system message if not using cached prompt
                api_params['system'] = system_message
                    
                # Only add tools if we have any
                if antr_tools:
                    api_params['tools'] = antr_tools
                
                # Make the API call
                result = self.client.messages.create(**api_params)
            except APIConnectionError as e:
                print("LLM call failed (APIConnectionError):", repr(e))
                raise ConnectionException(e)
            except (AuthenticationError, PermissionDeniedError) as e:
                print("LLM call failed (Auth):", repr(e))
                raise AuthorizationException(e)
            except InternalServerError as e:
                print("LLM call failed (500):", repr(e))
                raise ModelOverloadException(e)
            except RateLimitError as e:
                print("LLM call failed (RateLimit):", repr(e))
                raise RatelimitException(e)
            except BadRequestError as e:
                print("LLM call failed (BadRequest):", repr(e))
                raise BadRequestException(e)
            except Exception as e:
                print("LLM call failed (Unexpected):", repr(e))
                raise GeneralException(e)

            # Check for tool use in the response
            tool_use_blocks = [block for block in result.content if hasattr(block, 'type') and block.type == "tool_use"]
            
            if tool_use_blocks:
                # Create a new assistant message with the tool use blocks
                assistant_content = []
                tool_results = []
                
                for block in result.content:
                    if hasattr(block, 'type') and block.type == 'tool_use':
                        # Add tool use to the assistant's message
                        assistant_content.append({
                            "type": "tool_use",
                            "id": block.id,
                            "name": block.name,
                            "input": block.input
                        })
                        
                        # Execute the function
                        function = self.encapsulating_model.functions.get(block.name)
                        if not function:
                            raise ValueError(f"Function {block.name} not found in model's functions")
                        
                        function_result = function(**block.input)
                        
                        # Prepare tool result for the next user message
                        tool_results.append({
                            "type": "tool_result",
                            "tool_use_id": block.id,
                            "content": json.dumps(function_result) if not isinstance(function_result, (str, int, float, bool)) else str(function_result)
                        })
                    elif hasattr(block, 'text'):
                        assistant_content.append({"type": "text", "text": block.text})
                
                # Add the assistant's message with tool use to history
                self.messages.append({
                    "role": "assistant",
                    "content": assistant_content
                })
                
                # Add tool results as a new user message if there are any
                if tool_results:
                    self.messages.append({
                        "role": "user",
                        "content": tool_results
                    })
                
                # Continue the conversation with the tool results
                continue
            
            # If we get here, we have a final response with no tool use
            # Add the assistant's response to conversation history
            assistant_content = []
            for block in result.content:
                if hasattr(block, 'text'):
                    assistant_content.append({"type": "text", "text": block.text})
                elif hasattr(block, 'tool_use'):
                    assistant_content.append({
                        "type": "tool_use",
                        "id": block.tool_use.id,
                        "name": block.tool_use.name,
                        "input": block.tool_use.input
                    })
            
            if assistant_content:
                self.messages.append({
                    "role": "assistant",
                    "content": assistant_content
                })
            
            return result

    def cached_system_message(self) -> list[dict]:
        return [
                  {
                    "type": "text",
                    "text": self.system_message,
                  },
                  {
                    "type": "text",
                    "text": self.cached_prompt,
                    "cache_control": {"type": "ephemeral"}
                  }
                ]

    def token_count(self, text: str) -> int:
        message = create_anthropic_message('user', text)
        response = self.client.messages.count_tokens(model=self.model_name, messages=[message])
        return response.input_tokens


def transform_tools(tools: list[dict]) -> list[dict]:
    """
    Transforms tools into the format expected by Anthropic's API.
    """
    if not tools:
        return None
        
    type_mapping = {
        int: "integer",
        str: "string",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
    }
    
    transformed_tools = []
    
    for tool in tools:
        if "function" in tool:
            # This is a function tool
            tool_def = {
                "name": tool["function"].__name__,
                "description": tool.get("description", ""),
                "input_schema": {
                    "type": "object",
                    "properties": {},
                    "required": []
                }
            }
            
            # Add parameters to input_schema
            for param_name, param_type in tool.get("parameters", {}).items():
                tool_def["input_schema"]["properties"][param_name] = {
                    "type": type_mapping.get(param_type, "string"),
                    "description": param_name
                }
                
            # Add required parameters
            if "required_parameters" in tool:
                tool_def["input_schema"]["required"] = tool["required_parameters"]
                
            transformed_tools.append(tool_def)
    
    return transformed_tools if transformed_tools else None


def create_anthropic_message(role: str, prompt: str, images: ImageInput = None):

    content = []

    if images:
        for img in images:
            base64img = to_base64_image(img)
            mime_type = identify_image_format_from_base64(base64img)
            content += [
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": mime_type,
                        "data": base64img,
                    },
                }
            ]

    if prompt:
        content += [{"type": "text", "text": prompt}]

    return {"role": role, "content": content}


def extract_dict(text: str) -> dict:
    """ Extracts the first valid JSON dictionary from a string, even if it contains nested objects. """
    start = text.find("{")
    if start == -1:
        raise json.JSONDecodeError('No JSON object found in text', text, 0)

    depth = 0
    for i, ch in enumerate(text[start:], start=start):
        if ch == "{":
            depth += 1
        elif ch == "}":
            depth -= 1
            if depth == 0:
                return json.loads(text[start:i+1])
    raise json.JSONDecodeError('No valid JSON object found', text, start)



""" Implementation of the OpenAI Responses APImodels.

Feature table:
    - Async chat:       YES
    - Return JSON:      NO
    - Structured types: YES, via Pydantic  TODO: Add support for native Python types
    - Token count:      YES
    - Image support:    YES
    - Tool use:         YES
    - File input:       NOT YET IMPLEMENTED
    - Web search:       NOT YET IMPLEMENTED
"""
from __future__ import annotations
import base64
import json
import os
from io import BytesIO
from typing import Any, AsyncGenerator
import pydantic
from typing import Any, List, Tuple

from jsonschema import exceptions, validators, Draft202012Validator
import tiktoken
from PIL import Image
from dotenv import dotenv_values
from openai import OpenAI, APIConnectionError, \
    RateLimitError, APITimeoutError, AuthenticationError, PermissionDeniedError, BadRequestError

from justai.model.message import Message, ToolUseMessage
from justai.models.basemodel import ImageInput
from justai.tools.display import color_print, ERROR_COLOR
from justai.models.basemodel import BaseModel, DEFAULT_TIMEOUT, ConnectionException, AuthorizationException, \
    ModelOverloadException, RatelimitException, BadRequestException, GeneralException
from justai.tools.images import extract_images, to_base64_image, to_base64_data_uri, get_image_type


class OpenAIResponsesModel(BaseModel):
    def __init__(self, model_name: str, params: dict = None):
        params = params or {}
        system_message = f"You are {model_name}, a large language model trained by OpenAI."
        super().__init__(model_name, params, system_message)

        # Authentication
        api_key = params.get("OPENAI_API_KEY") or os.getenv("OPENAI_API_KEY") or dotenv_values()["OPENAI_API_KEY"]
        if not api_key:
            color_print("No OpenAI API key found. Create one at https://platform.openai.com/account/api-keys and " +
                        "set it in the .env file like OPENAI_API_KEY=here_comes_your_key.", color=ERROR_COLOR)

        # instructor.patch makes the OpenAI client compatible with structured output via response_model="
        # Works only for OpenAI models

        # Not sure if this works, or is needed, for the Responses API
        # self.client = instructor.patch(OpenAI(api_key=api_key))
        self.client = OpenAI(timeout=params.get('timeout', DEFAULT_TIMEOUT))

        # Diversions from the features that are supported or not supported by default
        self.supports_function_calling = True
        self.supports_image_generation = True

        self.last_response_id = None

    def prompt(self, prompt: str, images: list[ImageInput], tools, return_json: bool, response_format, _chat=False) \
            -> tuple[Any, int|None, int|None]:

        # if response_format and not issubclass(response_format, pydantic.BaseModel):
        #     raise NotImplementedError("OpenAI Responses API requires response_format to be a Pydantic model.")

        content = self.create_content(prompt, images)
        input_list = [{'role': 'system', 'content': self.system_message},
                      {'role': 'user', 'content': content}]
        tool_spec = self.create_tool_spec(tools)

        last_response_id = self.last_response_id if _chat else None

        for run in range(3):  # Max 3 function calls to prevent infinite loop
            try:
                if response_format and return_json:
                    assert is_valid_json_schema(response_format), "Response format should be a valid JSON Schema"
                    response = self.client.responses.create(model=self.model_name, input=input_list, tools=tool_spec,
                                                            text = {"format": {
                                                                "type": "json_schema",
                                                                "name": "response_format",
                                                                "strict": True,
                                                                "schema":response_format}},
                                                            previous_response_id=last_response_id)
                elif response_format:
                    assert isinstance(response_format, type) and issubclass(response_format, pydantic.BaseModel), \
                        'Response format should be a Pydantic model unless you specify return_json=True'
                    response = self.client.responses.parse(model=self.model_name, input=input_list, tools=tool_spec,
                                                           text_format=response_format,
                                                           previous_response_id=last_response_id)
                elif return_json:
                    response = self.client.responses.create(model=self.model_name, input=input_list, tools=tool_spec,
                                                            text = { "format": { "type": "json_object" } },
                                                            previous_response_id=last_response_id)
                else:
                    response = self.client.responses.create(model=self.model_name, input=input_list, tools=tool_spec,
                                                            previous_response_id=last_response_id)
            except APITimeoutError as e:
                raise ModelOverloadException(e)
            except APIConnectionError as e:
                raise ConnectionException(e)
            except (AuthenticationError, PermissionDeniedError) as e:
                raise AuthorizationException(e)
            except RateLimitError as e:
                raise RatelimitException(e)
            except BadRequestError as e:
                raise BadRequestException(e)
            except Exception as e:
                raise GeneralException(e)

            # Save the response id for subsequent requests
            self.last_response_id = response.id if _chat else None

            # Save function call outputs for subsequent requests
            function_call = None
            function_call_arguments = None
            input_list += response.output

            for item in response.output:
                if item.type == "function_call":
                    function_call = item
                    function_call_arguments = json.loads(item.arguments)

            if not function_call or run == 2:
                if response_format and isinstance(response_format, type) and issubclass(response_format, pydantic.BaseModel):
                    field = list(response.output_parsed.model_fields_set)[0]
                    output = getattr(response.output_parsed, field)
                elif return_json:
                    output = json.loads(response.output_text)
                else:
                    output = response.output_text
                return output, response.usage.input_tokens, response.usage.output_tokens

            # The rest of the function is to process the function call output
            # Numbering is from OpenAI's docs: https://platform.openai.com/docs/guides/function-calling#function-tool-example

            for tool in tools:
                if tool['function'].__name__ == function_call.name:
                    function = tool['function']
                    break
            else:
                raise ValueError(f"Function {function_call.name} not found")

            # 3. Execute the function logic for get_horoscope
            result = {function_call.name: function(*function_call_arguments.values())}

            # 4. Provide function call results to the model
            input_list.append(
                {
                    "type": "function_call_output",
                    "call_id": function_call.call_id,
                    "output": json.dumps(result),
                }
            )
        # Response looks like this:
        # [{
        #   "id": "resp_67ccd2bed1ec8190b14f964abc0542670bb6a6b452d3795b",
        #   "object": "response",
        #   "created_at": 1741476542,
        #   "status": "completed",
        #   "error": null,
        #   "incomplete_details": null,
        #   "instructions": null,
        #   "max_output_tokens": null,
        #   "model": "gpt-4.1-2025-04-14",
        #   "output": [
        #     {
        #       "type": "message",
        #       "id": "msg_67ccd2bf17f0819081ff3bb2cf6508e60bb6a6b452d3795b",
        #       "status": "completed",
        #       "role": "assistant",
        #       "content": [
        #         {
        #           "type": "output_text",
        #           "text": "In a peaceful grove beneath a silver moon, a unicorn named Lumina discovered a hidden pool that reflected the stars. As she dipped her horn into the water, the pool began to shimmer, revealing a pathway to a magical realm of endless night skies. Filled with wonder, Lumina whispered a wish for all who dream to find their own hidden magic, and as she glanced back, her hoofprints sparkled like stardust.",
        #           "annotations": []
        #         }
        #       ]
        #     }
        #   ],
        #   "parallel_tool_calls": true,
        #   "previous_response_id": null,
        #   "reasoning": {
        #     "effort": null,
        #     "summary": null
        #   },
        #   "store": true,
        #   "temperature": 1.0,
        #   "text": {
        #     "format": {
        #       "type": "text"
        #     }
        #   },
        #   "tool_choice": "auto",
        #   "tools": [],
        #   "top_p": 1.0,
        #   "truncation": "disabled",
        #   "usage": {
        #     "input_tokens": 36,
        #     "input_tokens_details": {
        #       "cached_tokens": 0
        #     },
        #     "output_tokens": 87,
        #     "output_tokens_details": {
        #       "reasoning_tokens": 0
        #     },
        #     "total_tokens": 123
        #   },
        #   "user": null,
        #   "metadata": {}
        # }]


    async def prompt_async(self, prompt: str, images: list[ImageInput]=None, _chat=False) -> AsyncGenerator[tuple[str, str], None]:

        content = self.create_content(prompt, images)
        input_ = [{"role": "user", "content": content}]

        last_response_id = self.last_response_id if _chat else None

        response = self.client.responses.create(model=self.model_name, input=input_, stream=True,
                                                previous_response_id=last_response_id)

        # Save the response id for subsequent requests
        self.last_response_id = response.id if _chat else None

        for event in response:
            if hasattr(event, 'delta'):
                yield event.delta, ''  # Second value is reasoning (not available for OpenAI)
        # Events can have different types:
        # event: response.created
        # data: {
        #     "type":"response.created",
        #     "response":{
        #         "id":"resp_67c9fdcecf488190bdd9a0409de3a1ec07b8b0ad4e5eb654",
        #         "object":"response",
        #         "created_at":1741290958,
        #         "status":"in_progress",
        #         "error":null,
        #         "incomplete_details":null,
        #         "instructions":"You are a helpful assistant.",
        #         "max_output_tokens":null,
        #         "model":"gpt-4.1-2025-04-14","
        #         output":[],"p
        #         arallel_tool_calls":true,
        #         "previous_response_id":null,
        #         "reasoning":{
        #             "effort":null,
        #             "summary":null},
        #         "store":true,
        #         "temperature":1.0,
        #         "text":{
        #             "format":{
        #                 "type":"text"}},
        #         "tool_choice":"auto",
        #         "tools":[],
        #         "top_p":1.0,
        #         "truncation":
        #         "disabled",
        #         "usage":null,
        #         "user":null,
        #         "metadata":{}}}
        #
        # event: response.in_progress
        # data: {
        #     "type":"response.in_progress",
        #     "response":{"
        #         id":"resp_67c9fdcecf488190bdd9a0409de3a1ec07b8b0ad4e5eb654",
        #         "object":"response",
        #         "created_at":1741290958,
        #         "status":"in_progress",
        #         "error":null,
        #         "incomplete_details":null,
        #         "instructions":"You are a helpful assistant.",
        #         "max_output_tokens":null,
        #         "model":"gpt-4.1-2025-04-14",
        #         "output":[],
        #         "parallel_tool_calls":true,
        #         "previous_response_id":null,
        #         "reasoning":{
        #             "effort":null,
        #             "summary":null},
        #         "store":true,
        #         "temperature":1.0,
        #         "text":{
        #             "format":{
        #                 "type":"text"}},
        #         "tool_choice":"auto",
        #         "tools":[],
        #         "top_p":1.0,
        #         "truncation":"disabled",
        #         "usage":null,
        #         "user":null,
        #         "metadata":{}}}
        #
        # event: response.output_item.added
        # data: {
        #     "type":"response.output_item.added",
        #     "output_index":0,
        #     "item":{
        #         "id":"msg_67c9fdcf37fc8190ba82116e33fb28c507b8b0ad4e5eb654",
        #         "type":"message","status":"
        #         in_progress","
        #         role":"assistant",
        #         "content":[]}}
        #
        # event: response.content_part.added
        # data: {
        #     "type":"response.content_part.added",
        #     "item_id":"msg_67c9fdcf37fc8190ba82116e33fb28c507b8b0ad4e5eb654",
        #     "output_index":0,
        #     "content_index":0,
        #     "part":{
        #         "type":"output_text",
        #         "text":"",
        #         "annotations":[]}}
        #
        # event: response.output_text.delta
        # data: {
        #     "type":"response.output_text.delta",
        #     "item_id":"msg_67c9fdcf37fc8190ba82116e33fb28c507b8b0ad4e5eb654",
        #     "output_index":0,
        #     "content_index":0,
        #     "delta":"Hi"}
        #
        # ...
        #
        # event: response.output_text.done
        # data: {"type":"response.output_text.done","item_id":"msg_67c9fdcf37fc8190ba82116e33fb28c507b8b0ad4e5eb654","output_index":0,"content_index":0,"text":"Hi there! How can I assist you today?"}
        #
        # event: response.content_part.done
        # data: {"type":"response.content_part.done","item_id":"msg_67c9fdcf37fc8190ba82116e33fb28c507b8b0ad4e5eb654","output_index":0,"content_index":0,"part":{"type":"output_text","text":"Hi there! How can I assist you today?","annotations":[]}}
        #
        # event: response.output_item.done
        # data: {"type":"response.output_item.done","output_index":0,"item":{"id":"msg_67c9fdcf37fc8190ba82116e33fb28c507b8b0ad4e5eb654","type":"message","status":"completed","role":"assistant","content":[{"type":"output_text","text":"Hi there! How can I assist you today?","annotations":[]}]}}
        #
        # event: response.completed
        # data: {"type":"response.completed","response":{"id":"resp_67c9fdcecf488190bdd9a0409de3a1ec07b8b0ad4e5eb654","object":"response","created_at":1741290958,"status":"completed","error":null,"incomplete_details":null,"instructions":"You are a helpful assistant.","max_output_tokens":null,"model":"gpt-4.1-2025-04-14","output":[{"id":"msg_67c9fdcf37fc8190ba82116e33fb28c507b8b0ad4e5eb654","type":"message","status":"completed","role":"assistant","content":[{"type":"output_text","text":"Hi there! How can I assist you today?","annotations":[]}]}],"parallel_tool_calls":true,"previous_response_id":null,"reasoning":{"effort":null,"summary":null},"store":true,"temperature":1.0,"text":{"format":{"type":"text"}},"tool_choice":"auto","tools":[],"top_p":1.0,"truncation":"disabled","usage":{"input_tokens":37,"output_tokens":11,"output_tokens_details":{"reasoning_tokens":0},"total_tokens":48},"user":null,"metadata":{}}}

    def chat(self, prompt: str, images: list[ImageInput], tools, return_json: bool, response_format) \
             -> tuple[Any, int|None, int|None]:
        return self.prompt(prompt, images, tools, return_json, response_format, _chat=True)


    async def chat_async(self, prompt: str, images: list[ImageInput]) -> AsyncGenerator[tuple[str, str], None]:
        async for chunk in self.prompt_async(prompt, images, _chat=True):
            yield chunk
               
    @staticmethod
    def create_content(prompt, images: list[ImageInput]) -> list[dict]:
        content = [{"type": "input_text", "text": prompt}]

        if images:
            for image in images:
                # Always convert to base64 data URI to avoid URL download issues
                # Some servers (like Wikipedia) block OpenAI's download attempts
                image_url = to_base64_data_uri(image)
                content += [{"type": "input_image", "image_url": image_url}]
        return content

    @staticmethod
    def create_tool_spec(tools: list[dict]) -> list[dict]:
        if not tools:
            return []

        type_mapping = {
            int: "integer",
            str: "string",
            float: "number",
            bool: "boolean",
            list: "array",
            dict: "object"
        }
        tool_spec = []
        for tool in tools:
            tool_spec += [{
                "type": "function",
                "name": tool["function"].__name__,
                "description": tool['description'],
                "parameters": {
                    "type": "object",
                    "properties": {
                        param_name: {
                            "type": type_mapping.get(_type, "string"),  # Default to string if type not in map
                            "description": param_name,  # Or a more descriptive text if available
                        }
                        for param_name, _type in tool['parameters'].items()
                    },
                    "required": tool['required_parameters'] or [],
                }
            }]
        return tool_spec

    @staticmethod
    def transform_messages(messages: list[Message]) -> list[dict]:
        transformed_messages = []

        for message in messages:
            msg = {"role": message.role}

            # Handle tool messages (function calls and their results)
            # Todo: tool use is geen onderdeel meer van message maar staat in de Model class.
            if isinstance(message, ToolUseMessage):
                if message.role == 'assistant' and 'function_to_call' in message.tool_use:
                    # This is a function call from the assistant
                    msg["content"] = None
                    msg["tool_calls"] = [{
                        "id": message.tool_use.get('call_id', 'call_' + str(hash(str(message.tool_use)))),
                        "type": "function",
                        "function": {
                            "name": message.tool_use['function_to_call'],
                            "arguments": json.dumps(message.tool_use['function_parameters'])
                        }
                    }]
                elif message.role == 'tool':
                    # This is a function result
                    function_result = message.tool_use.get('function_result', '')
                    if not isinstance(function_result, str):
                        function_result = json.dumps(function_result)
                    msg["content"] = function_result
                    msg["tool_call_id"] = message.tool_use.get('call_id', '')
                    msg["name"] = message.tool_use.get('function_to_call', '')
            # Handle regular messages
            else:
                if message.images:
                    content = [{"type": "text", "text": message.content or ""}]
                    for image in message.images:
                        content.append({
                            "type": "image_url",
                            "image_url": {'url': to_base64_data_uri(image)}
                        })
                    msg["content"] = content
                else:
                    msg["content"] = message.content or ""

            transformed_messages.append(msg)

        return transformed_messages

    def token_count(self, text: str) -> int:
        """ Returns the number of tokens in a string. """
        try:
            encoding = tiktoken.encoding_for_model(self.model_name)
        except KeyError:
            # Fall back to cl100k_base encoding for newer models not yet in tiktoken
            encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(text))


    def generate_image(self, prompt, images: ImageInput, options: dict = None):

        # Responses API call met tool "image_generation"
        client = OpenAI()
        input_structure = [
            {
                "role": "user",
                "content": [
                    {"type": "input_text", "text": prompt}
                ],
            }
        ]
        if images:
            for image in images:
                if get_image_type(image) == "image_url":
                    image_dict = {"type": "input_image", "image_url": image}
                else:
                    image_dict = {"type": "input_image", "image_data": to_base64_image(image)}
                input_structure[0]["content"] += [image_dict]

        resp = client.responses.create(
            model=self.model_name,
            input=input_structure,
            tools=[{"type": "image_generation"}],
        )

        images_b64 = extract_images(resp)

        # Decode base64 -> PIL image
        raw = base64.b64decode(images_b64[0])
        img = Image.open(BytesIO(raw))
        return img


def is_valid_json_schema(schema: Any) -> Tuple[bool, List[str]]:
    """
    Check whether a Python object is a *valid JSON Schema* (meta-schema validation).

    It auto-detects the draft via `$schema` when present and falls back to Draft 2020-12.
    Returns (is_valid, errors) where errors is a list of human-readable messages.

    Parameters
    ----------
    schema : Any
        The candidate JSON Schema as a Python dict (or JSON-loaded structure).

    Returns
    -------
    Tuple[bool, List[str]]
        - True and [] when the input is a valid JSON Schema.
        - False and a list of error messages when invalid.

    Notes
    -----
    - This validates the *schema itself* against the appropriate meta-schema.
    - It does not validate an instance/document *against* the schema.
    """

    def _format_iter_error(err: exceptions.ValidationError) -> str:
        """Human-friendly error message for iter_errors results."""
        # Build a JSON Pointer–like path for clarity
        path = "/" + "/".join(map(str, err.path)) if err.path else "(root)"
        schema_path = "#/" + "/".join(map(str, err.schema_path)) if err.schema_path else "#"
        return f"[at {path}] {err.message}  (schema path: {schema_path})"


    def _format_schema_error(err: exceptions.SchemaError) -> str:
        """Format SchemaError raised by check_schema()."""
        path = "/" + "/".join(map(str, err.path)) if getattr(err, "path", None) else "(root)"
        schema_path = "#/" + "/".join(map(str, err.schema_path)) if getattr(err, "schema_path", None) else "#"
        base = f"[at {path}] {err.message}  (schema path: {schema_path})"
        if err.context:
            # Include nested context messages (useful for 'oneOf', 'anyOf' diagnostics)
            ctx = "; ".join(_format_iter_error(c) for c in err.context)  # type: ignore[arg-type]
            return f"{base} | context: {ctx}"
        return base

    error_messages: List[str] = []

    # 1) Pick the best validator class for the provided schema (uses `$schema` if present)
    try:
        ValidatorClass = validators.validator_for(schema, default=Draft202012Validator)  # type: ignore[attr-defined]
    except Exception as e:
        return False, [f"Unable to choose validator for schema (is it a dict?): {e!r}"]

    # 2) Fast sanity check (raises on the first structural problem)
    try:
        ValidatorClass.check_schema(schema)
    except exceptions.SchemaError as e:
        # We still continue to collect *all* errors below for a complete report
        error_messages.append(_format_schema_error(e))

    # 3) Full meta-schema validation to gather all issues
    try:
        meta_validator = ValidatorClass(ValidatorClass.META_SCHEMA)  # type: ignore[arg-type]
        errors = sorted(meta_validator.iter_errors(schema), key=lambda err: (list(err.path), err.message))
        if errors:
            for err in errors:
                error_messages.append(_format_iter_error(err))
    except Exception as e:
        # If meta-validation itself fails, report that clearly
        error_messages.append(f"Meta-validation failed: {e!r}")

    return (len(error_messages) == 0), error_messages



""" Implementation of the Google models.
https://ai.google.dev/gemini-api/docs/migrate

Feature table:
    - Async chat:       YES (1)
    - Return JSON:      YES
    - Structured types: YES, via Python type definition
    - Token counter:    YES
    - Image support:    YES 
    - Tool use:         NO (not yet)

Supported parameters:
    max_output_tokens= 400,
    top_k= 2,
    top_p= 0.5,
    temperature= 0.5,
    response_mime_type= 'application/json',
    stop_sequences= ['\n'],
    seed=42,

(1) In contrast to Model.chat, Model.chat_async cannot return json and does not return input and output token counts

"""
import json
import os
from io import BytesIO
from typing import Any, AsyncGenerator

from PIL import Image
from dotenv import dotenv_values
from google import genai

from justai.model.message import Message
from justai.model.model import ImageInput
from justai.models.basemodel import BaseModel, DEFAULT_TIMEOUT
from justai.tools.display import ERROR_COLOR, color_print
from justai.tools.images import to_pil_image

class GoogleModel(BaseModel):

    def __init__(self, model_name: str, params: dict = None):
        params = params or {}
        system_message = f"You are {model_name}, a large language model trained by Google."
        super().__init__(model_name, params, system_message)

        # Authentication
        api_key = params.get("GEMINI_API_KEY") or params.get("GOOGLE_API_KEY") or os.getenv("GEMINI_API_KEY") or \
                  os.getenv("GOOGLE_API_KEY") or dotenv_values()["GEMINI_API_KEY"] or dotenv_values()["GOOGLE_API_KEY"]
        if not api_key:
            color_print("No Google API key found. Create one at https://aistudio.google.com/app/apikey and " +
                        "set it in the .env file like GOOGLE_API_KEY=here_comes_your_key.", color=ERROR_COLOR)

        # Client (Google uses milliseconds for timeout)
        timeout_ms = int(params.get('timeout', DEFAULT_TIMEOUT) * 1000)
        http_options = genai.types.HttpOptions(timeout=timeout_ms)
        self.client = genai.Client(api_key=api_key, http_options=http_options)
        self.chat_session = None  # Google uses this to keep track of the chat

        # Diversions from the features that are supported or not supported by default
        self.supports_function_calling = True
        self.supports_automatic_function_calling = True
        self.supports_image_generation = True

    def prompt(self, prompt: str, images: ImageInput, tools: list, return_json: bool, response_format) -> str | object:
        if isinstance(images, str):
            images = [images]
        opened_images = [to_pil_image(img) for img in images] if images else []
        if opened_images:
            prompt = [prompt] + opened_images
        if tools and isinstance(tools[0], dict):
            tools = [tool['function'] for tool in tools]
        config = genai.types.GenerateContentConfig(system_instruction=self.system_message, tools=tools,
                                                   **self.model_params)
        if return_json:
            config.response_mime_type = "application/json"
        if response_format:
            config.response_mime_type = "application/json"
            config.response_schema = response_format
        response = self.client.models.generate_content(model=self.model_name, contents=prompt,
                                                       config=config)
        return convert_to_justai_response(response, return_json or response_format)

    def chat(self, prompt: str, images: ImageInput, tools: list, return_json: bool, response_format) \
            -> tuple[Any, int|None, int|None]:

        if return_json:
            raise NotImplementedError('google_model.chat does not support return_json. Use prompt() instead')
        if response_format:
            raise NotImplementedError('google_model.chat does not support response_format. Use prompt() instead')
        if images:
            raise NotImplementedError('google_model.chat does not support images. Use prompt() instead')

        if not self.chat_session:
            self.chat_session = self.client.chats.create(model=self.model_name)
        response = self.chat_session.send_message(message=prompt)
        return convert_to_justai_response(response, return_json)

    async def prompt_async(self, prompt: str, images: list[ImageInput] = None) -> AsyncGenerator[tuple[str, str], None]:
        if images:
            raise NotImplementedError('google_model. ..._async does not support images. Use prompt() instead')

        stream = await self.client.aio.models.generate_content_stream(
            model=self.model_name,
            contents=prompt
        )
        async for chunk in stream:
            if chunk.text:
                yield chunk.text, ''

    async def chat_async(self, prompt: str, images: list[ImageInput] = None) -> AsyncGenerator[tuple[str, str], None]:
        async for chunk in self.prompt_async(prompt, images):
            yield chunk

    def token_count(self, text: str) -> int:
        response = self.client.models.count_tokens(model=self.model_name, contents=text)
        return response.total_tokens

    def generate_image(self, prompt, images: ImageInput, options: dict = None):
        images = [to_pil_image(img) for img in images] if images else []

        # Build config from options if provided
        config = None
        if options:
            config_params = {}
            if 'aspect_ratio' in options:
                config_params['aspect_ratio'] = options['aspect_ratio']
            if 'number_of_images' in options:
                config_params['number_of_images'] = options['number_of_images']
            if config_params:
                config = genai.types.GenerateContentConfig(**config_params)

        response = self.client.models.generate_content(
            model=self.model_name,
            contents=images + [prompt],
            config=config,
        )

        for part in response.candidates[0].content.parts:
            if part.text is not None:
                print(part.text)
            elif part.inline_data is not None:
                image = Image.open(BytesIO(part.inline_data.data))
                return image


def convert_to_justai_response(response, return_json):
    input_token_count = response.usage_metadata.prompt_token_count
    output_token_count = (response.usage_metadata.candidates_token_count or 0) + \
                         (response.usage_metadata.thoughts_token_count or 0)
    result = response.text if not return_json else response.parsed if response.parsed else json.loads(response.text)
    return result, input_token_count, output_token_count



from justai.models.basemodel import BaseModel


class ModelFactory:
    @staticmethod
    def create(model_name: str, **kwargs) -> BaseModel:
        if model_name.startswith("openrouter/"):
            from justai.models.openrouter_models import OpenRouterModel
            model_name = model_name.split("/", 1)[1]
            assert '/' in model_name, "Model name should be in the format 'openrouter/provider/modelname'"
            return OpenRouterModel(model_name, params=kwargs)
        elif model_name.startswith("gpt") or model_name.startswith("o1") or model_name.startswith("o3"):
            from justai.models.openai_responses import OpenAIResponsesModel
            return OpenAIResponsesModel(model_name, params=kwargs)
        elif model_name.endswith(".gguf"):
            from justai.models.gguf_models import GuffModel
            return GuffModel(model_name, params=kwargs)
        elif model_name.startswith("claude"):
            from justai.models.anthropic_models import AnthropicModel
            return AnthropicModel(model_name, params=kwargs)
        elif model_name.startswith("gemini"):
            from justai.models.google_models import GoogleModel
            return GoogleModel(model_name, params=kwargs)
        elif model_name.startswith("grok"):
            from justai.models.xai_models import XAIModel
            return XAIModel(model_name, params=kwargs)
        elif model_name.startswith("deepseek"):
            from justai.models.deepseek_models import DeepSeekModel
            return DeepSeekModel(model_name, params=kwargs)
        elif model_name.startswith("sonar"):
            from justai.models.perplexity_models import PerplexityModel
            return PerplexityModel(model_name, params=kwargs)
        elif model_name.startswith("reve"):
            from justai.models.reve_models import ReveModel
            return ReveModel(model_name, params=kwargs)
        else:
            raise ValueError(f"Model {model_name} not supported")



from pathlib import Path
from importlib.metadata import version, PackageNotFoundError

from justai.model.model import Model
from justai.tools.prompts import get_prompt, set_prompt_file, add_prompt_file


def _get_version():
    try:
        __version__ = version(__name__)
    except PackageNotFoundError:
        with open(Path(__file__).parent / "pyproject.toml", "r") as f:
            for line in f:
                if line.startswith("version ="):
                    return line.split('"')[1]
            raise RuntimeError("Unable to find version")


__version__ = _get_version()

# Use like this
# from importlib.metadata import version
# print(version("justai"))

if __name__ == '__main__':
    # Onderstaande om de voorkomen dat import optimizer ze leeg gooit
    a = Model
    g = get_prompt
    s = set_prompt_file
    apf = add_prompt_file
